==> cbra-0.46.0.tar.gz: cbra-0.46.0/cbra/ext/oauth2/server.py <==
"""Declares :class:`OpenAuthorizationServer`."""
import copy
import typing
import types

import ckms
import fastapi
import ckms.jose.exceptions
import ckms.jose.models
from ckms.jose import PayloadCodec
from ckms.jose.models import JSONWebKeySet
from fastapi.responses import RedirectResponse
from starlette.middleware.base import BaseHTTPMiddleware

import cbra
from cbra.ext import ioc
from .exceptions import Error
from .exceptions import ErrorModel
from .iopenauthorizationserver import IOpenAuthorizationServer
from .iclientrepository import IClientRepository
from .istorage import IStorage
from .isubjectrepository import ISubjectRepository
from .openauthorizationendpoint import OpenAuthorizationEndpoint
from .servermetadata import ServerMetadata
from .tokenissuer import TokenIssuer
from .tokenrequesthandler import TokenRequestHandler


class OpenAuthorizationServer(fastapi.APIRouter, IOpenAuthorizationServer):
    app: cbra.Application
    storage: typing.Optional[typing.Type[IStorage]]
    clients: typing.Type[IClientRepository]
    subjects: typing.Type[ISubjectRepository]
    keychain: ckms.Keychain
    jwks: JSONWebKeySet
    enable_request: bool = False
    container: ioc.Provider
    default_signing_key: typing.Optional[str] = None
    default_encryption_key: typing.Optional[str] = None

    @classmethod
    def fromsettings(
        cls,
        settings: types.ModuleType,
        **kwargs: typing.Any
    ) -> 'OpenAuthorizationServer':
        server = cls(
            enable_request=getattr(settings, 'OAUTH2_ENABLE_PAR', False),
            **kwargs
        )
        server.register_keys( # type: ignore
            default_encryption_key=settings.DEFAULT_ENCRYPTION_KEY,
            default_signing_key=settings.DEFAULT_SIGNING_KEY,
            keys=settings.OAUTH2_SERVER_KEYS,
        )
        return server

    def __init__(
        self,
        *,
        clients: typing.Type[IClientRepository],
        subjects: typing.Type[ISubjectRepository],
        storage: typing.Optional[typing.Type[IStorage]] = None,
        keychain: typing.Optional[ckms.Keychain] = None,
        authorize: typing.Optional[OpenAuthorizationEndpoint] = None,
        exchange: typing.Optional[OpenAuthorizationEndpoint] = None,
        token: typing.Optional[typing.Type[TokenRequestHandler]] = None,
        token_factory: typing.Optional[typing.Type[TokenIssuer]] = TokenIssuer,
        enable_request: bool = False,
        **kwargs: typing.Any
    ):
        self.authorization = None
        self.clients = clients
        self.subjects = subjects
        self.container = ioc.Provider()
        self.keychain = keychain or ckms.Keychain()
        self.authorize_url = kwargs.pop('authorize_url', '/authorize')
        self.enable_request = enable_request
        self.exchange_url = kwargs.pop('exchange_url', '/exchange')
        self.jwks_url = kwargs.pop('jwks_url', '/jwks')
        self.jwks = JSONWebKeySet()
        self.token = token
        self.token_url = kwargs.pop('token_url', '/token')
        self.storage = storage
        self.metadata = ServerMetadata(
            grant_types_supported=TokenIssuer.grant_types_supported,
            token_endpoint_auth_methods_supported=["none"]
        )

        on_startup = kwargs.setdefault('on_startup', [])
        on_shutdown = kwargs.setdefault('on_shutdown', [])

        on_startup.append(self.on_startup)
        on_shutdown.append(self.on_shutdown)

        super().__init__(**kwargs)
        self.add_api_route(
            path=self.jwks_url,
            endpoint=self.get_jwks,
            name='oauth2.jwks',
            summary="JSON Web Key Set (JWKS) endpoint",
            response_description=(
                "A JSON Web Key Set (JWKS) holding the public keys used by "
                "the authorization server."
            )
        )
        if exchange:
            self.add_api_route(
                path=self.exchange_url,
                endpoint=OpenAuthorizationEndpoint(self, exchange),
                name='oauth2.exchange',
                summary="Token exchange endpoint",
                response_model=exchange.response_model,
                response_model_exclude_unset=True,
                methods=["POST"],
                responses={
                    400: {'model': ErrorModel}
                }
            )

    def register_keys(
        self,
        default_signing_key: str,
        default_encryption_key: str,
        keys: typing.Dict[str, typing.Dict[str, typing.Any]]
    ) -> None:
        """Registers keys for use with the :class:`Server`.
        
        The `default_signing_key` and the `default_encryption_key`
        specify the keys used to sign and decrypt messages sent by
        the server, such as JSON Web Signatures (JWS) or JSON Web
        Encryptio (JWE).
        """
        if default_signing_key not in keys:
            raise ValueError(
                "The `keys` mapping must contain `default_signing_key`."
            )
        if default_encryption_key not in keys:
            raise ValueError(
                "The `keys` mapping must contain `default_encryption_key`."
            )

        # Ensure that all keys in the mapping have the `oauth2` tag.
        keys = copy.deepcopy(keys)
        for kid, spec in keys.items():
            tags = spec.setdefault('tags', [])
            if 'oauth2' not in tags:
                tags.append('oauth2')
            if kid == default_encryption_key:
                spec['use'] = 'enc'
            if kid == default_signing_key:
                spec['use'] = 'sig'

        self.default_encryption_key = default_encryption_key
        self.default_signing_key = default_signing_key
        self.keychain.register(keys)

    def add_to_app(self, app: cbra.Application, base_path: str = '/'):
        """Adds a :class:`OpenAuthorizationServer` to an ASGI application."""
        self.app = app
        app.add_middleware(BaseHTTPMiddleware, dispatch=self.on_request)
        app.inject('cbra.ext.oauth2.ClientRepository', self.clients)
        app.inject('cbra.ext.oauth2.SubjectRepository', self.subjects)
        app.inject(
            name='cbra.ext.oauth2.ServerCodec',
            value=PayloadCodec(signer=self.keychain, decrypter=self.keychain)
        )
        app.inject(
            name='cbra.ext.oauth2.TransientStorage',
            value=self.storage
        )

        if self.token is not None:
            self.add_api_route(
                path=self.token_url,
                endpoint=OpenAuthorizationEndpoint(self, self.token),
                name='oauth2.token',
                summary="Token endpoint",
                methods=["POST"],
                response_model=self.token.response_model,
                response_model_exclude_unset=True,
                responses={
                    400: {'model': ErrorModel},
                    422: {}
                },
                response_description=(
                    "A JSON datastructure containing the access token as specified in "
                    "RFC 6749."
                )
            )

        app.include_router(
            router=self,
            prefix=str.rstrip(base_path, '/'),
            tags=["OAuth 2.0/OpenID Connect"]
        )
        app.add_api_route(
            path='/.well-known/oauth-authorization-server',
            endpoint=self.metadata_endpoint,
            tags=["OAuth 2.0/OpenID Connect"],
            name='oauth2.metadata',
            summary="Authorization server metadata",
            default_error_responses=False,
            response_model=ServerMetadata,
            response_description=(
                "A JSON datastructure describing the authorization server (see RFC 8414)."
            )
        )
        app.add_api_route(
            path='/.well-known/openid-configuration',
            endpoint=self.metadata_redirect,
            include_in_schema=False
        )

        app.add_exception_handler(Error, self.handle_oauth2_error) # type: ignore

    async def handle_oauth2_error(self, request: fastapi.Request, exc: Error):
        return await exc.as_response(request=request)

    async def on_request(
        self,
        request: fastapi.Request,
        call_next: typing.Callable[
            ...,
            typing.Coroutine[typing.Any, typing.Any, fastapi.Response]
        ]
    ) -> fastapi.Response:
        """Setup the request context for the authorization server."""
        request.scope['oauth2'] = self
        return await call_next(request)

    async def on_startup(self):
        await self.keychain.setup()
        self.jwks = JSONWebKeySet.fromtags(self.keychain, ['oauth2'])
        self.app.inject('cbra.ext.oauth2.ServerJWKS', self.jwks)

    async def on_shutdown(self):
        """Hook to perform cleanup operations after shutdown."""
        pass

    async def get_jwks(self) -> ckms.jose.models.JSONWebKeySet:
        return self.jwks.dict(exclude_defaults=True)

    async def metadata_endpoint(self, request: fastapi.Request) -> ServerMetadata:
        metadata = {
            **self.metadata.dict(),
            'issuer': f'{request.url.scheme}://{request.url.netloc}',
            'token_endpoint': request.url_for('oauth2.token') if self.token else None,
            'jwks_uri': request.url_for('oauth2.jwks')
        }
        return ServerMetadata(**metadata)

    async def metadata_redirect(self, request: fastapi.Request):
        return RedirectResponse(
            request.url_for('oauth2.metadata'),
            status_code=303
        )

==> cbra-0.50.0.tar.gz: cbra-0.50.0/cbra/ext/oauth2/server.py <==
"""Declares :class:`OpenAuthorizationServer`."""
import copy
import typing
import types

import ckms
import fastapi
import ckms.jose.exceptions
import ckms.jose.models
from ckms.jose import PayloadCodec
from ckms.jose.models import JSONWebKeySet
from fastapi.responses import RedirectResponse
from starlette.middleware.base import BaseHTTPMiddleware

import cbra
from cbra.ext import ioc
from .exceptions import Error
from .exceptions import ErrorModel
from .iopenauthorizationserver import IOpenAuthorizationServer
from .iclientrepository import IClientRepository
from .istorage import IStorage
from .isubjectrepository import ISubjectRepository
from .openauthorizationendpoint import OpenAuthorizationEndpoint
from .servermetadata import ServerMetadata
from .tokenissuer import TokenIssuer
from .tokenrequesthandler import TokenRequestHandler


class OpenAuthorizationServer(fastapi.APIRouter, IOpenAuthorizationServer):
    app: cbra.Application
    storage: typing.Optional[typing.Type[IStorage]]
    clients: typing.Type[IClientRepository]
    subjects: typing.Type[ISubjectRepository]
    keychain: ckms.Keychain
    enable_request: bool = False
    container: ioc.Provider
    default_signing_key: typing.Optional[str] = None
    default_encryption_key: typing.Optional[str] = None

    @classmethod
    def fromsettings(
        cls,
        settings: types.ModuleType,
        **kwargs: typing.Any
    ) -> 'OpenAuthorizationServer':
        server = cls(
            enable_request=getattr(settings, 'OAUTH2_ENABLE_PAR', False),
            **kwargs
        )
        server.register_keys( # type: ignore
            default_encryption_key=settings.DEFAULT_ENCRYPTION_KEY,
            default_signing_key=settings.DEFAULT_SIGNING_KEY,
            keys=settings.OAUTH2_SERVER_KEYS,
        )
        return server

    @property
    def jwks(self) -> JSONWebKeySet:
        return self.app.get_jwks()

    def __init__(
        self,
        *,
        clients: typing.Type[IClientRepository],
        subjects: typing.Type[ISubjectRepository],
        storage: typing.Optional[typing.Type[IStorage]] = None,
        authorize: typing.Optional[OpenAuthorizationEndpoint] = None,
        exchange: typing.Optional[OpenAuthorizationEndpoint] = None,
        token: typing.Optional[typing.Type[TokenRequestHandler]] = None,
        token_factory: typing.Optional[typing.Type[TokenIssuer]] = TokenIssuer,
        enable_request: bool = False,
        default_signing_key: str = 'sig',
        default_encryption_key: str = 'enc',
        **kwargs: typing.Any
    ):
        self.authorization = None
        self.clients = clients
        self.default_encryption_key = default_encryption_key
        self.default_signing_key = default_signing_key
        self.subjects = subjects
        self.container = ioc.Provider()
        self.authorize_url = kwargs.pop('authorize_url', '/authorize')
        self.enable_request = enable_request
        self.exchange_url = kwargs.pop('exchange_url', '/exchange')
        self.token = token
        self.token_url = kwargs.pop('token_url', '/token')
        self.storage = storage
        self.metadata = ServerMetadata(
            grant_types_supported=TokenIssuer.grant_types_supported,
            token_endpoint_auth_methods_supported=["none"]
        )

        on_startup = kwargs.setdefault('on_startup', [])
        on_shutdown = kwargs.setdefault('on_shutdown', [])

        on_startup.append(self.on_startup)
        on_shutdown.append(self.on_shutdown)
        super().__init__(**kwargs)

    def add_to_app(self, app: cbra.Application, base_path: str = '/'):
        """Adds a :class:`OpenAuthorizationServer` to an ASGI application."""
        self.app = app
        self.keychain = app.keychain
        app.add_middleware(BaseHTTPMiddleware, dispatch=self.on_request)
        app.inject('cbra.ext.oauth2.ClientRepository', self.clients)
        app.inject('cbra.ext.oauth2.SubjectRepository', self.subjects)

        assert self.keychain is not None
        app.inject(
            name='cbra.ext.oauth2.ServerCodec',
            value=PayloadCodec(signer=self.keychain, decrypter=self.keychain)
        )
        app.inject(
            name='cbra.ext.oauth2.TransientStorage',
            value=self.storage
        )

        if self.token is not None:
            self.add_api_route(
                path=self.token_url,
                endpoint=OpenAuthorizationEndpoint(self, self.token),
                name='oauth2.token',
                summary="Token endpoint",
                methods=["POST"],
                response_model=self.token.response_model,
                response_model_exclude_unset=True,
                responses={
                    400: {'model': ErrorModel},
                    422: {}
                },
                response_description=(
                    "A JSON datastructure containing the access token as specified in "
                    "RFC 6749."
                )
            )

        app.include_router(
            router=self,
            prefix=str.rstrip(base_path, '/'),
            tags=["OAuth 2.0/OpenID Connect"]
        )
        app.add_api_route(
            path='/.well-known/oauth-authorization-server',
            endpoint=self.metadata_endpoint,
            tags=["OAuth 2.0/OpenID Connect"],
            name='oauth2.metadata',
            summary="Authorization server metadata",
            default_error_responses=False,
            response_model=ServerMetadata,
            response_description=(
                "A JSON datastructure describing the authorization server (see RFC 8414)."
            )
        )
        app.add_api_route(
            path='/.well-known/openid-configuration',
            endpoint=self.metadata_redirect,
            include_in_schema=False
        )

        app.add_exception_handler(Error, self.handle_oauth2_error) # type: ignore

    async def handle_oauth2_error(self, request: fastapi.Request, exc: Error):
        return await exc.as_response(request=request)

    async def on_request(
        self,
        request: fastapi.Request,
        call_next: typing.Callable[
            ...,
            typing.Coroutine[typing.Any, typing.Any, fastapi.Response]
        ]
    ) -> fastapi.Response:
        """Setup the request context for the authorization server."""
        request.scope['oauth2'] = self
        return await call_next(request)

    async def on_startup(self):
        await self.keychain.setup()
        self.app.inject('cbra.ext.oauth2.ServerJWKS', self.jwks)

    async def on_shutdown(self):
        """Hook to perform cleanup operations after shutdown."""
        pass

    async def metadata_endpoint(self, request: fastapi.Request) -> ServerMetadata:
        metadata = {
            **self.metadata.dict(),
            'issuer': f'{request.url.scheme}://{request.url.netloc}',
            'token_endpoint': request.url_for('oauth2.token') if self.token else None,
            'jwks_uri': request.url_for('metadata.jwks')
        }
        return ServerMetadata(**metadata)

    async def metadata_redirect(self, request: fastapi.Request):
        return RedirectResponse(
            request.url_for('oauth2.metadata'),
            status_code=303
        )
